"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
// Copyright 2019 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
const crypto = require("crypto");
const retry = require("p-retry");
const zlib = require("zlib");
const dockerio_1 = require("./auth/dockerio");
const gcr_1 = require("./auth/gcr");
const credentials_helper_1 = require("./credentials-helper");
const image_specifier_1 = require("./image-specifier");
const packer = require("./packer");
const pending_1 = require("./pending");
const registry_1 = require("./registry");
// expose plain registry client.
var registry_2 = require("./registry");
exports.RegistryClient = registry_2.RegistryClient;
class Image {
    // queue all pending actions so i can Promise.all them in save before saving
    // manifest
    constructor(imageSpecifier, targetImage, options) {
        this.clients = {};
        this.syncedBaseImage = false;
        this.options = options || {};
        if (typeof targetImage !== 'string') {
            this.options = this.options || targetImage;
            targetImage = undefined;
        }
        this.image = image_specifier_1.parse(imageSpecifier);
        this.targetImage = image_specifier_1.parse(targetImage || imageSpecifier);
        this.pending = pending_1.pending();
        // setup default client.
        // if the to and from match host and namespace we'll make one write client.
        const readOnly = this.authKey(this.image) !== this.authKey(this.targetImage);
        // optimistic client loading. if it error's the user will be able to get the
        // error the next time they grab a client or try to save.
        this.client(this.image, !readOnly)
            .catch(() => {
            // we can ignore the unhandled rejection here because we cache
            // this promise and return it when they try to get the client.
        });
        this.imageData = false;
    }
    // returns the part of the config object you care about for things like
    // entrypoint and env.
    async getImageConfig() {
        const imageData = await this.getImageData();
        return imageData.config.config;
    }
    async addLayer(digest, uncompressedDigest, size, urls) {
        const imageData = await this.getImageData();
        let layerMediaType = 'application/vnd.oci.image.layer.v1.tar+gzip';
        if (imageData.manifest.mediaType.indexOf('docker') > -1) {
            layerMediaType = 'application/vnd.docker.image.rootfs.diff.tar.gzip';
        }
        const layerResult = { mediaType: layerMediaType, digest, size, urls };
        imageData.manifest.layers.push(layerResult);
        imageData.config.rootfs.diff_ids.push(uncompressedDigest);
        return Object.assign({}, layerResult, { uncompressedDigest });
    }
    async removeLayer(digest) {
        // edit the config and manifest returned from getImageData
        const imageData = await this.getImageData();
        const layers = imageData.manifest.layers;
        let found;
        layers.forEach((layerData, i) => {
            if (layerData.digest === digest) {
                found = i;
            }
        });
        if (found !== undefined) {
            layers.splice(found, 1);
            imageData.config.rootfs.diff_ids.splice(found, 1);
            return true;
        }
        return false;
    }
    // "./myfiles", "/workspace"
    // "localDirectory", "imageDirectory"
    addFiles(dir, targetDir, options) {
        // dir,target,options
        // dir,options
        // {dir:target,....},options
        if (typeof targetDir === 'string') {
            if (typeof dir !== 'string') {
                // addFiles({"apples":"oranges"},"pears")
                throw new Error('specifying a target directory name when the dir is an object of name:target doesn\'t make sense. try addFiles({dir:target})');
            }
            dir = { [targetDir]: dir };
        }
        else if (targetDir) {
            // options!
            options = targetDir;
        }
        // have to wrap in promise because the tar stream can emit error out of band
        let p = new Promise(async (resolve, reject) => {
            const tarStream = packer.pack(dir, options);
            tarStream.on('error', (e) => reject(e));
            const gzip = zlib.createGzip();
            const uncompressedHash = crypto.createHash('sha256');
            tarStream.on('data', (buf) => {
                uncompressedHash.update(buf);
            });
            tarStream.pipe(gzip);
            const client = await this.client(this.targetImage, true);
            const result = await client.upload(gzip);
            const uncompressedDigest = 'sha256:' + uncompressedHash.digest('hex');
            resolve(await this.addLayer(result.digest, uncompressedDigest, result.contentLength));
        });
        p = this.pending.track(p);
        return p;
    }
    async getImageData() {
        if (!this.imageData) {
            this.imageData = await this.loadImageData();
            this.originalManifest =
                JSON.parse(JSON.stringify(this.imageData.manifest));
        }
        return this.imageData;
    }
    async loadImageData(image) {
        image = (image ? image : this.image);
        const client = await this.client(image);
        const manifest = await client.manifest(image.tag || 'latest');
        const configBlob = await client.blob(manifest.config.digest) + '';
        const config = JSON.parse(configBlob);
        return { manifest, config };
    }
    client(_image, write) {
        let image;
        if (typeof _image === 'string') {
            image = image_specifier_1.parse(_image);
        }
        else {
            // typescript!!!
            image = _image;
        }
        image = (image ? image : this.image);
        const scope = write ? 'push,pull' : 'pull';
        let key = [image.registry, image.namespace, image.image].join(',');
        const writeKey = key + ',push,pull';
        const readKey = key + ',pull';
        // default to most permissive cached client even if it doesn't match scope.
        if (this.clients[writeKey]) {
            return this.clients[writeKey];
        }
        else if (!write && this.clients[readKey]) {
            return this.clients[readKey];
        }
        key += ',' + scope;
        const promiseOfClient = exports.auth(image, scope, this.options.auth || {}).then((registryAuth) => {
            const registryClient = new registry_1.RegistryClient(image.registry, this.nameSpacedImageName(image), registryAuth);
            return registryClient;
        });
        this.clients[key] = promiseOfClient;
        return promiseOfClient;
    }
    async save(tags, options) {
        const targetImage = this.targetImage;
        const client = await this.client(targetImage, true);
        const imageData = await this.getImageData();
        tags = tags || [targetImage.tag || 'latest'];
        options = options || {};
        await this.syncBaseImage(options);
        await Promise.all(this.pending.active());
        if (options.Cmd || this.Cmd) {
            imageData.config.config.Cmd = options.Cmd || this.Cmd;
        }
        if (options.Env || this.Env) {
            imageData.config.config.Env =
                imageData.config.config.Env.concat(options.Env || this.Env || []);
        }
        if (options.WorkingDir || this.WorkingDir) {
            imageData.config.config.WorkingDir =
                options.WorkingDir || this.WorkingDir;
        }
        // are all layers done uploading?
        const uploadResult = await client.upload(Buffer.from(JSON.stringify(imageData.config)));
        imageData.manifest.config.digest = uploadResult.digest;
        imageData.manifest.config.size = uploadResult.contentLength;
        // put the manifest once per tag
        return Promise
            .all(tags.filter((v) => !!v).map((tag) => {
            return client.manifestUpload(encodeURIComponent(tag), imageData.manifest);
        }))
            .then((results) => {
            return results[0];
        });
    }
    authKey(image) {
        if (image.registry.indexOf('gcr.io')) {
            // should use same auth if in same google cloud project.
            return [image.registry, image.namespace].join(',');
        }
        return [image.registry, image.namespace, image.image].join(',');
    }
    // verify that every layer is in the target and copy missing layers from base
    // to target
    async sync(options) {
        options =
            Object.assign({ copyRemoteLayers: true, ignoreExists: false }, options);
        // ensure image data has been loaded.
        await this.getImageData();
        // use base manifest for sync
        const manifest = this.originalManifest;
        if (!manifest) {
            throw new Error('get image data failed to populate originalManifest somehow.');
        }
        const client = await this.client(this.image);
        const targetClient = await this.client(this.targetImage, true);
        // check that every layer in source image is present in target registry
        const copies = [];
        manifest.layers.forEach((layer) => {
            // note: support *not* copying nondistributable layers by default once
            // it's supported by gcr.
            if (!options.copyRemoteLayers && layer.urls) {
                return;
            }
            const p = targetClient.blobExists(layer.digest).then((exists) => {
                if (!exists || options.ignoreExists) {
                    // TODO if they are the same registry but different namespaces try
                    // mount first.
                    const action = () => {
                        return new Promise(async (resolve, reject) => {
                            const stream = await client.blob(layer.digest, true);
                            // if the stream ends without the correct byte length
                            let bytes = 0;
                            stream.on('data', (b) => {
                                bytes += b.length;
                            });
                            stream.on('end', () => {
                                if (bytes !== layer.size) {
                                    reject(Error('failed to get all of the bytes from the blob stream.'));
                                }
                            });
                            stream.on('error', (err) => {
                                reject(err);
                            });
                            try {
                                resolve(await targetClient.upload(stream, layer.size, layer.digest));
                            }
                            catch (e) {
                                console.error('error syncing layer ' + layer.digest +
                                    ' to target registry:\n' + e);
                                reject(e);
                            }
                        });
                    };
                    return retry(action, { retries: 3 }).then(() => true);
                }
                return true;
            });
            copies.push(p);
        });
        const all = Promise.all(copies);
        this.pending.track(all);
        return all;
    }
    async syncBaseImage(options) {
        const sameRegistry = this.image.registry === this.targetImage.registry;
        const sameProject = this.image.namespace === this.targetImage.namespace;
        if (sameRegistry && sameProject) {
            return;
        }
        if (this.syncedBaseImage) {
            return;
        }
        this.syncedBaseImage = true;
        return this.sync(options);
    }
    nameSpacedImageName(image) {
        image = (image ? image : this.image);
        return (image.namespace ? image.namespace + '/' : '') + image.image;
    }
}
exports.Image = Image;
// the ordinary auth mechanism that might work on any docker registry is not
// implemented. https://docs.docker.com/registry/spec/auth/token/
exports.auth = async (imageArg, scope, options) => {
    // todo: distinguish better between when we should try creds helpers vs only
    // built in.
    let image;
    if (typeof imageArg === 'string') {
        image = image_specifier_1.parse(imageArg);
    }
    else {
        image = imageArg;
    }
    try {
        if (image.registry.indexOf('gcr.io') > -1) {
            return await gcr_1.handler(image, scope, options ? options[image.registry + '/' + image.namespace] ||
                options['registry'] || options['gcr.io'] || {} :
                {});
        }
        else if (image.registry.indexOf('docker.io') > -1) {
            // dockerhub requires you log in every 5 minutes.
            // we'll always try to get a new token for the user with the provided auth
            // data
            return await dockerio_1.handler(image, scope, options ? options['docker.io'] : undefined);
        }
    }
    catch (e) {
        console.error('gcr or docker.io auth threw.\n' + e +
            '\n falling back to cred helpers.');
    }
    // if the user provided auth options
    if (options) {
        const checked = ['gcr.io', 'docker.io'];
        let providedAuth;
        Object.keys(options).forEach((s) => {
            if (!checked.includes(s)) {
                if (image.registry.indexOf(s) > -1) {
                    // we have auth provided directly for this registry.
                    providedAuth = options[s];
                }
            }
        });
        // if we dont have an auth helper we assume the user has passed valid
        // credentials.
        if (providedAuth) {
            return providedAuth;
        }
    }
    const credHelpers = new credentials_helper_1.DockerCredentialHelpers();
    const res = await credHelpers.auth(image.registry);
    return res;
};
exports.pack = packer.pack;
// expose CustomFile to pass in image.addFiles
exports.CustomFile = packer.CustomFile;
//# sourceMappingURL=index.js.map